/* prefetch.h - prefetch module interfaces and definitions */

#ifndef PREFETCH_H
#define PREFETCH_H

#include <stdio.h>

#include "host.h"
#include "misc.h"
#include "machine.h"
#include "memory.h"
#include "stats.h"
#include "cache.h"

/* The various trigger types */

typedef enum {
    Cache_Miss,           /* A miss in the cache */     
    Cache_Hit,            /* A hit in the cache */
    Memory_Access,        /* An access to memory by an instruction */
    PC_Update,            /* Program counter is updated */
    No_event              /* Dummy trigger */
} trigger_type_t;

/* This enum handles whare prefetching triggers happen */

typedef enum {
		Cache_IL1,             /* Event happened in the D-I1 cache */
		Cache_DL1,             /* Event happened in the D-L1 cache */
		Cache_IL2,             /* Event happened in the I-L2 cache */
		Cache_DL2,             /* Event happened in the D-L2 cache */
		DRAM,                  /* Event happened in DRAM */
		None                   /* When location doesn't matter - eg PC_Update */
} location_t;

/* The main prefetching structure */

typedef struct {
  trigger_type_t type;     /* What happened */
  location_t location;     /* Where something did happen */
  md_addr_t address;       /* Adress of memory access */
  tick_t time;             /* Time of access */
} prefetch_trigger_t;

/* States for the rpt table */
/* Defined in the paper by Chen and Baer */

typedef enum {
  Initial,
  Transient,
  Steady,
  No_prediction
} rpt_state_t;

/* RPT table structure */
typedef struct {
  md_addr_t pc;             /* Adress of loading instruction */
  md_addr_t prev_addr;      /* Previous adress loaded by instruction */
  int stride;               /* Recorded stride */
  rpt_state_t state;        /* Current state */
  tick_t atime;             /* Time of last access (in ticks) */
} rpt_entry_t;

/* AVD prefetcher as in the paper:
 * "Adress-Value Delta (AVD) Prediction: Increasing the Effectiveness of Runahead
 * Execution by Exploiting Regular Memory Allocation Patterns" by
 * Onur Mutlu, Hyesoon Kim and Uale N. Patt
 */

/* AVD table structure */
typedef struct {
  md_addr_t pc;             /* Adress of loading instruction */
  int avd;                  /* Calculated delta */
  tick_t atime;             /* Time of last access (in ticks) */
  int confidence;           /* Confidence in prediction */
} avd_entry_t;


/* Prototype declarations */
md_addr_t set_last_PC_value(md_addr_t PC);

int register_prefetch_location(location_t location);

void register_prefetch_degree(int degree);

void register_prefetch_target(trigger_type_t target);

int register_prefetch_algorithm(void (*algorithm)(prefetch_trigger_t));

void set_return_latency(int latency);

int prefetch_init(struct cache_t *il1, struct cache_t *il2, struct cache_t *dl1, struct cache_t *dl2,int c_size,int t_size, struct mem_t *memory);

void prefetch(md_addr_t adress, struct cache_t *target, time_t now);

void no_prefetch(prefetch_trigger_t trigger);

void sequential_prefetch(prefetch_trigger_t trigger);

void delta_correlation_prefetch(prefetch_trigger_t trigger);

void czone_delta_correlation_prefetch(prefetch_trigger_t trigger);

void rpt_prefetch(prefetch_trigger_t trigger);

void avd_prefetch(prefetch_trigger_t trigger);

void stream_prefetch(prefetch_trigger_t trigger);

int process_prefetch_trigger(trigger_type_t type, location_t location, md_addr_t address, tick_t now);

#endif /* PREFETCH_H */
